use std::collections::HashMap;

use dioxus::prelude::*;

pub(crate) trait Form {
    type FieldName: Eq + std::hash::Hash + 'static;

    fn set(&mut self, field_name: Self::FieldName, value: String);

    fn get(&self, field_name: &Self::FieldName) -> Option<String>;
}

pub(crate) struct FormState<FieldName> {
    state: Box<dyn Form<FieldName = FieldName>>,
    errors: HashMap<FieldName, Vec<Box<dyn std::error::Error>>>,
}

impl<FieldName> FormState<FieldName> {
    fn new<S: 'static>(state: S) -> Self
    where
        S: Form<FieldName = FieldName>,
    {
        Self {
            state: Box::new(state),
            errors: Default::default(),
        }
    }
}

impl<FieldName> FormState<FieldName>
where
    FieldName: Eq + std::hash::Hash + 'static,
{
    pub(crate) fn add_error<E: 'static>(&mut self, name: FieldName, error: E) -> &mut Self
    where
        E: std::error::Error,
    {
        self.errors.entry(name).or_default().push(Box::new(error));
        self
    }

    pub(crate) fn clear_errors(&mut self) -> &mut Self {
        self.errors.clear();
        self
    }

    fn set(&mut self, name: FieldName, value: String) {
        self.state.set(name, value);
    }

    fn get_value(&self, name: &FieldName) -> Option<String> {
        self.state.get(name)
    }

    fn get_error(&self, name: &FieldName) -> Option<&[Box<dyn std::error::Error>]> {
        self.errors.get(name).and_then(|vec| {
            if vec.is_empty() {
                None
            } else {
                Some(vec.as_slice())
            }
        })
    }
}

#[inline_props]
#[allow(non_snake_case)]
fn InputErrors<'a, FieldName>(cx: Scope, name: &'a FieldName) -> Element
where
    FieldName: Eq + std::hash::Hash + 'static,
{
    let errors = use_shared_state::<FormState<FieldName>>(cx);

    if let Some(errors) = errors {
        let errors = errors.read();
        if let Some(errors) = errors.get_error(name) {
            return cx.render(rsx! {
                div {
                    class: "text-input--errors",
                    errors.iter().map(|error| rsx! {
                        span {
                            class: "text-input--errors--error",
                            "{error}"
                        }
                    })
                }
            });
        }
    }

    None
}

#[inline_props]
#[allow(non_snake_case)]
pub(crate) fn TextInput<'a, FieldName>(cx: Scope, name: FieldName, label: Element<'a>) -> Element
where
    FieldName: Clone + Eq + std::hash::Hash + std::fmt::Display + 'static,
{
    cx.render(rsx! {
        Input { kind: "text", name: name.clone(), label: label.clone() }
    })
}

#[inline_props]
#[allow(non_snake_case)]
pub(crate) fn PasswordInput<'a, FieldName>(
    cx: Scope,
    name: FieldName,
    label: Element<'a>,
) -> Element
where
    FieldName: Clone + Eq + std::hash::Hash + std::fmt::Display + 'static,
{
    cx.render(rsx! {
        Input { kind: "password", name: name.clone(), label: label.clone() }
    })
}

#[inline_props]
#[allow(non_snake_case)]
pub(crate) fn Input<'a, FieldName>(
    cx: Scope,
    kind: &'static str,
    name: FieldName,
    label: Element<'a>,
) -> Element
where
    FieldName: Clone + Eq + std::hash::Hash + std::fmt::Display + 'static,
{
    cx.render(rsx! {
        label {
            class: "text-input",
            span {
                class: "text-input--label",
                label,
            },
            match use_shared_state::<FormState<FieldName>>(cx) {
                Some(values) => {
                    let value = values.read().get_value(&name).unwrap_or_else(|| String::new());

                    rsx! {
                        input {
                            name: "{name}",
                            r#type: "{kind}",
                            value: "{value}",
                            oninput: move |event| values.write().set(name.clone(), event.value.clone()),
                        },
                    }
                },
                None => rsx! {
                    input {
                        name: "{name}",
                        r#type: "{kind}",
                    },
                },
            },
            InputErrors { name: name, }
        }
    })
}

impl<T> Form for UseRef<T>
where
    T: Form,
{
    type FieldName = T::FieldName;

    fn set(&mut self, field_name: Self::FieldName, value: String) {
        self.with_mut(|form| form.set(field_name, value));
    }

    fn get(&self, field_name: &Self::FieldName) -> Option<String> {
        self.with(|form| form.get(field_name))
    }
}

pub(crate) fn use_form<FieldName, State>(
    cx: &ScopeState,
    form: UseRef<State>,
) -> UseSharedState<'_, FormState<FieldName>>
where
    State: Form<FieldName = FieldName> + Default + 'static,
{
    use_shared_state_provider(cx, move || FormState::<FieldName>::new(form));

    use_shared_state(cx).expect("Just set up")
}
