use std::{hash::Hash, net::SocketAddr};

use axum::{
    extract::WebSocketUpgrade,
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use dioxus::prelude::*;
use dioxus_liveview::LiveViewPool;
use dioxus_router::{Link, Redirect, Route, Router};

mod form;

use form::{use_form, PasswordInput, TextInput};

#[derive(Clone, Debug, Default)]
struct Form {
    username: String,
    password: String,
    password_confirmation: String,
}

#[derive(Clone, Debug, PartialEq, Eq, Hash)]
enum Field {
    Username,
    Password,
    ConfirmPassword,
}

#[derive(Debug)]
enum Error {
    UsernameBlank,
    UsernameTooShort,
    PasswordBlank,
    PasswordTooShort,
    PasswordMismatch,
}

impl form::Form for Form {
    type FieldName = Field;

    fn set(&mut self, field_name: Self::FieldName, value: String) {
        match field_name {
            Field::Username => {
                self.username = value;
            }
            Field::Password => {
                self.password = value;
            }
            Field::ConfirmPassword => {
                self.password_confirmation = value;
            }
        }
    }

    fn get(&self, field_name: &Self::FieldName) -> Option<String> {
        Some(match field_name {
            Field::Username => self.username.clone(),
            Field::Password => self.password.clone(),
            Field::ConfirmPassword => self.password_confirmation.clone(),
        })
    }
}

impl std::fmt::Display for Field {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Username => write!(f, "username"),
            Self::Password => write!(f, "password"),
            Self::ConfirmPassword => write!(f, "confirm-password"),
        }
    }
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::UsernameBlank => write!(f, "Username is required"),
            Self::UsernameTooShort => write!(f, "Username is too short"),
            Self::PasswordBlank => write!(f, "Password is required"),
            Self::PasswordTooShort => write!(f, "Password is too short"),
            Self::PasswordMismatch => write!(f, "Passwords must match"),
        }
    }
}

impl std::error::Error for Error {}

#[allow(non_snake_case)]
fn Home(cx: Scope) -> Element {
    let form = use_ref(cx, Form::default);
    let form_state = use_form(cx, form.clone());

    cx.render(rsx! {
        form {
            onsubmit: move |_| {
                let form = form.read();
                let mut form_state = form_state.write();
                form_state.clear_errors();

                if form.username.is_empty() {
                    form_state.add_error(Field::Username, Error::UsernameBlank);
                }
                if form.username.len() < 5 {
                    form_state.add_error(Field::Username, Error::UsernameTooShort);
                }

                if form.password.is_empty() {
                    form_state.add_error(Field::Password, Error::PasswordBlank);
                }
                if form.password.len() < 8 {
                    form_state.add_error(Field::Password, Error::PasswordTooShort);
                }

                if form.password != form.password_confirmation {
                    form_state.add_error(Field::ConfirmPassword, Error::PasswordMismatch);
                }

                println!("{:?}", form);
            },

            TextInput {
                name: Field::Username,
                label: cx.render(rsx! { "Username" }),
            },

            PasswordInput {
                name: Field::Password,
                label: cx.render(rsx! { "Password" }),
            },

            PasswordInput {
                name: Field::ConfirmPassword,
                label: cx.render(rsx! { "Confirm Password" }),
            },

            input {
                r#type: "submit",
                value: "Sign Up"
            }
        }
    })
}

fn application(cx: Scope) -> Element {
    cx.render(rsx! {
        Router {
            nav {
                div {
                    Link { to: "/posts", "Home" },
                },
            },
            Route { to: "/posts", Home {} },
            Redirect { from: "", to: "/posts" },
        }
    })
}

async fn spa(addr: SocketAddr) -> Html<String> {
    Html(format!(
        r#"
    <!DOCTYPE html>
    <html>
    <head>
        <title>Dioxus LiveView with Axum</title>
        <style>
        * {{
            box-sizing: border-box;
        }}
        body {{
            margin: 0;
            min-height: 100vh;
            background-color: #333;
            color: #f5f5f5;
            font-family: sans;
        }}
        </style>
    </head>
    <body> <div id="main"></div> </body>
    {glue}
    </html>
    "#,
        // Create the glue code to connect to the WebSocket on the "/ws" route
        glue = dioxus_liveview::interpreter_glue(&format!("ws://{addr}/ws"))
    ))
}

async fn websocket(ws: WebSocketUpgrade, view: LiveViewPool) -> impl IntoResponse {
    ws.on_upgrade(move |socket| async move {
        // When the WebSocket is upgraded, launch the LiveView with the app component
        _ = view
            .launch(dioxus_liveview::axum_socket(socket), application)
            .await;
    })
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let addr: SocketAddr = ([127, 0, 0, 1], 3030).into();

    let view = dioxus_liveview::LiveViewPool::new();

    let app = Router::new()
        // The root route contains the glue code to connect to the WebSocket
        .route("/", get(move || spa(addr)))
        // The WebSocket route is what Dioxus uses to communicate with the browser
        .route("/ws", get(move |sock| websocket(sock, view)));

    println!("Listening on http://{addr}");

    axum::Server::bind(&addr.to_string().parse()?)
        .serve(app.into_make_service())
        .await?;

    Ok(())
}
